/*
  centerpos.cpp
  Copyright (C) 2008-2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

//#define DEBUG

#include "qgmap.h"
#include "centerpos.h"
#include "pos.h"
#include "layeredbase.h"
#include "mainwidget.h"
#include "inputdialog.h"
#include "loc_phs.h"
#include "qgconf.h"

#include <qapplication.h>
#include <qclipboard.h>

#define MOVE_INTERVAL 100
#define MOVE_STEPS 8

CenterPos::CenterPos(LayeredBase *parent, const char *name) :
	LayeredWidget(parent, name)
{
	move_event_counter = 0;
	move_step = -1;

	/* connect signals */
	connect(this, SIGNAL(positionChanged(Pos &)),
			parent, SIGNAL(positionChanged(Pos &)));
	connect(parent, SIGNAL(setPosition(Pos &)),
			this, SLOT(setPosition(Pos &)));
	connect(parent, SIGNAL(movePosition(Pos &)),
			this, SLOT(movePosition(Pos &)));
	connect(parent, SIGNAL(zoomChanged(int)),
			this, SLOT(setZoom(int)));

}

const Pos &CenterPos::getPosition()
{
	return center;
}

void CenterPos::setPosition(Pos &pos)
{
	Enter();
	pos.normalize();
	pos.clipY();
	if(center == pos)
		return;

	center = pos;
	emit positionChanged(center);
}

void CenterPos::movePosition(Pos &pos)
{
	Enter();

	move_to = pos;
	move_step = 0;

	killTimers();
	startTimer(MOVE_INTERVAL);
}

void CenterPos::setZoom(int zoom)
{
	if(center.getZoom() == zoom)
		return;

	if(center.zoomTo(zoom))
		emit positionChanged(center);
}

void CenterPos::scroll(int x, int y)
{
	setPosition(center.offset(x, y));
}

void CenterPos::zoomup(int z)
{
	if(center.zoomup(z))
		emit positionChanged(center);
}

bool CenterPos::wheelEvent( QWheelEvent *e )
{
	Enter();
	if (move_step >= 0)
		return false;

	zoomup(e->delta()<0?1:-1);
	return true;
}

bool CenterPos::mouseMoveEvent( QMouseEvent *e )
{
	Enter();

	if (move_step >= 0)
		return false;

	QPoint pos = click_pos - e->pos();

	scroll(pos.x(), pos.y());

	click_pos = e->pos();
	return true;
}

bool CenterPos::mousePressEvent( QMouseEvent *e )
{
	Enter();

	if (move_step >= 0)
		return false;

	if(e->button() != LeftButton)
		return false;

	click_pos = e->pos();
	return true;
}

bool CenterPos::mouseReleaseEvent( QMouseEvent *e )
{
	Enter();

	if (move_step >= 0)
		return false;

	if(e->button() != LeftButton)
		return false;

	repaint();
	return true;
}

bool CenterPos::mouseDoubleClickEvent(QMouseEvent *e)
{
	Enter();

	if (move_step >= 0)
		return false;

	Pos to = center.offset(e->x() - this->width() / 2,
						   e->y() - this->height() / 2);

	movePosition(to);
	return true;
}

bool CenterPos::keyPressEvent ( QKeyEvent * e )
{
	int shift = (e->state() & ShiftButton)?SHIFT_MUL:1;

	if (move_step >= 0)
		return false;

	switch(e->key()){
	case Key_Left:
		scroll(-KEY_SCROLL*shift, 0);
		return true;
	case Key_Right:
		scroll(KEY_SCROLL*shift, 0);
		return true;
	case Key_Up:
		scroll(0, -KEY_SCROLL*shift);
		return true;
	case Key_Down:
		scroll(0, KEY_SCROLL*shift);
		return true;
	case Key_I:
		zoomup(-1);
		return true;
	case Key_O:
		zoomup(1);
		return true;
	case Key_J:
		jumpLocation();
		return true;
	case Key_P:
		getPHSLocation();
		return true;

	// copy & paste
	case Key_C:
		if(e->state() & ControlButton)  // Fn+C も Ctrl+C と同じ
			copyPosition();
		break;
	case Key_V:
		if(e->state() & ControlButton)  // Fn+V も Ctrl+V と同じ
			pastePosition();
		break;

	// Save & Load marker pos
	case Key_0:
	case Key_1:
	case Key_2:
	case Key_3:
	case Key_4:
	case Key_5:
	case Key_6:
	case Key_7:
	case Key_8:
	case Key_9:
		if(e->state() & ShiftButton) {
			savePosition(e->key() - Key_0);
		} else {
			loadPosition(e->key() - Key_0);
		}
		return true;

#ifndef QWS	// for PC
	case Key_Exclam:
		savePosition(1);
		return true;
	case Key_QuoteDbl:
		savePosition(2);
		return true;
	case Key_NumberSign:
		savePosition(3);
		return true;
	case Key_Dollar:
		savePosition(4);
		return true;
	case Key_Percent:
		savePosition(5);
		return true;
	case Key_Ampersand:
		savePosition(6);
		return true;
	case Key_Apostrophe:
		savePosition(7);
		return true;
	case Key_ParenLeft:
		savePosition(8);
		return true;
	case Key_ParenRight:
		savePosition(9);
		return true;
	case Key_AsciiTilde:
		savePosition(0);
		return true;
#endif // for PC
	}
	return false;
}

void CenterPos::showEvent()
{
	resize(parentWidget()->size());
}

void CenterPos::savePosition(int i)
{
	QString key;
	QString info_str;
	QGConf *conf = (QGConf *)lookupObject(this, "QGConf", "config");

	key = tr("Pos%1").arg(i);
	if(conf)
		conf->writeEntry(key, center.serialize());

	info_str = QString::fromUtf8("Pos%1に保存").arg(i);
	((MainWidget *)parentWidget())->dispInfo(info_str);
}

void CenterPos::loadPosition(int i)
{
	Enter();
	QString key, str;
	QString info_str;
	Pos tmp;

	key.sprintf("Pos%d", i);
	QGConf *conf = (QGConf *)lookupObject(this, "QGConf", "config");
	if(conf){
		str = conf->readEntry(key, tmp.serialize());
		tmp.deserialize(str);
	}

	setPosition(tmp);

	info_str = QString::fromUtf8("Pos%1を読込").arg(i);
	((MainWidget *)parentWidget())->dispInfo(info_str);
}

void CenterPos::jumpLocation()
{
	InputDialog dlg(parentWidget(), tr("Jump Location (Lat,Lon[,zoom])"), TRUE);
	QString str;

	str.sprintf("%f,%f", center.getLat(), center.getLon());
	dlg.setText(str);

	if(dlg.exec() == QDialog::Accepted){
		Pos pos = center;
		if(pos.setPos(dlg.getText())){
			setPosition(pos);
		}
	}
}

void CenterPos::copyPosition()
{
	QClipboard *cb = QApplication::clipboard();

	cb->setText(center.getPos());
	((MainWidget *)parentWidget())->dispInfo(
		QString::fromUtf8("現在位置をクリップボードへ保存しました"));
}

void CenterPos::pastePosition()
{
	QClipboard *cb = QApplication::clipboard();
	Pos pos = center;

	if(pos.setPos(cb->text())){
		setPosition(pos);
	}else{
		((MainWidget *)parentWidget())->dispInfo(
			QString::fromUtf8("クリップボードの位置情報が不正です"));
	}
}

void CenterPos::getPHSLocation()
{
	LocPHS phs(parentWidget());
	Pos pos = center;

	if(!phs.getLocation(pos))
		return;

	setPosition(pos);
	((MainWidget *)parentWidget())->dispInfo(
		QString::fromUtf8("PHSから位置情報を取得しました"));
}

void CenterPos::timerEvent(QTimerEvent *e UNUSED)
{
	Enter();
	killTimers();

	Pos tmp;

	if (++move_step < MOVE_STEPS) {
		int x = move_to.getX() - center.getX();
		int y = move_to.getY() - center.getY();
		if ((x / 2) != 0)
			x = x / 2;
		if ((y / 2) != 0)
			y = y / 2;
		tmp = center.offset(x, y);
	}else{
		tmp = move_to;
	}

	if (tmp == move_to)
		move_step = MOVE_STEPS;

	setPosition(tmp);
}

bool CenterPos::paintEvent(QPaintEvent *e UNUSED, QPixmap &scr UNUSED)
{
	Enter();

	if(move_step < 0)
		return false;

	if(move_step < MOVE_STEPS){
		startTimer(MOVE_INTERVAL);
	}else{
		move_step = -1;
	}
	return false;
}

void CenterPos::request(QString const &req, QString const &arg)
{
	int ev = arg.toInt();
	if (req != "menu")
		return;

	if (ev >= EV_JUMP_MARKER_0 && ev <= EV_JUMP_MARKER_9)
		loadPosition(ev - EV_JUMP_MARKER_0);

	if (ev >= EV_SAVE_MARKER_0 && ev <= EV_SAVE_MARKER_9)
		savePosition(ev - EV_SAVE_MARKER_0);

	switch (ev) {
	case EV_JUMP:
		jumpLocation();
		break;
	case EV_GET_PHS_POS:
		getPHSLocation();
		break;
	case EV_COPY_LOCATION:
		copyPosition();
		break;
	case EV_PASTE_LOCATION:
		pastePosition();
		break;
	case EV_ZOOM_IN:
		zoomup(-1);
		break;
	case EV_ZOOM_OUT:
		zoomup(1);
		break;
	}
}
