/*
  freehandregion.cpp
  Copyright (C) 2008-2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

//#define DEBUG

#include <qwidget.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qstring.h>
#include <qpainter.h>
#include <qarray.h>
#include <qpointarray.h>
#include <qapplication.h>
#include <qclipboard.h>
#include "qgmap.h"
#include "mainwidget.h"
#include "freehandregion.h"
#include "pos.h"
#include "layeredbase.h"
#include "freehanddialog.h"
#include "regionmap.h"
#include "mainwindow.h"

#include <stdio.h>
#include <stdlib.h>

FreehandRegion::FreehandRegion( LayeredBase *parent, const char *name) :
	LayeredWidget(parent, name)
{
	Enter();
	is_button = FALSE;
	mode = NONE;

	canvas.resize(0,0);
	saved_screen.resize(0,0);

	/* connect signals */
	connect(parent, SIGNAL(positionChanged(Pos &)),
			this, SLOT(setPosition(Pos &)));
}

bool FreehandRegion::paintEvent(QPaintEvent *e UNUSED, QPixmap &scr)
{
	if(isHidden())
		return false;

	Enter();

	if(mode == NONE){
		return false;
	}
	if(mode == FETCH){
		fetchScreen(scr);
		mode = DRAW;
		clearParray();
	}

	bitBlt(&scr, QPoint(0,0), &canvas, canvas.rect());

	// draw polyline
	QPainter p(&scr);
	p.setPen(QPen(Qt::red,3));
	p.drawPolyline(points);

	return true;
}

bool FreehandRegion::mousePressEvent(QMouseEvent *e)
{
	Enter();
	if(mode == NONE)
		return false;

	is_button = TRUE;
	points.resize(0);
	old = e->pos();
	addPoint(e->pos());
	return true;
}

bool FreehandRegion::mouseReleaseEvent(QMouseEvent *e UNUSED)
{
	Enter();
	if(mode == NONE)
		return false;

	QPainter p(&canvas);
	p.setPen(QPen(Qt::red,3));
	p.setBrush(Qt::red);
	p.drawPolygon(points);

	int i = parray.size();
	parray.resize(i+1);
	parray.insert(i, new QPointArray(points.copy()));
	points.resize(0);
	repaint();

	is_button = false;
	return false;
}

bool FreehandRegion::mouseMoveEvent(QMouseEvent *e)
{
	QPoint tmp;
	int dist;

	Enter();
	if(mode == NONE)
		return false;

	if(!is_button)
		return false;

	tmp = e->pos() - old;
	dist = abs(tmp.x())+abs(tmp.y());
	if(dist < 3)
		return true;

	addPoint(e->pos());
	old=e->pos();
	repaint();
	return true;
}

bool FreehandRegion::keyPressEvent ( QKeyEvent * e )
{
	if(mode == NONE){
		if(e->key() == Key_D){
			mode_on();
			return true;
		}
		return false;
	}

	switch(e->key()){
	case Key_D:
		mode_off();
		break;

	case Key_Backspace:
		popParray();
		redrawCanvas();
		repaint();
		break;

	case Key_Return:
		if(parray.size()==0){
			mode_off();
			break;
		}

		if(save_region()){
			mode_off();
		}
	}
	return true;
}

void FreehandRegion::showEvent()
{
	Enter();
	QWidget *parent = parentWidget();
	int width = parent->width();
	int height = parent->height();

	setGeometry(0, 0, width, height);
}

void FreehandRegion::setPosition(Pos &pos)
{
	center = pos;
}

void FreehandRegion::addPoint(const QPoint &point)
{
	int index = points.size();
	points.resize(index+1);
	points.setPoint(index, point);
}

void FreehandRegion::clearParray()
{
	while(parray.size()>0)
		popParray();
}

void FreehandRegion::popParray()
{
	int i = parray.size()-1;
	if(i<0)
		return;

	delete parray[i];
	parray.resize(i);
}

void FreehandRegion::redrawCanvas()
{
	bitBlt(&canvas, QPoint(0,0), &saved_screen, canvas.rect());

	QPainter p(&canvas);
	p.setPen(QPen(Qt::red,3));
	p.setBrush(Qt::red);

	int i;
	for(i=0; i<(int)parray.size(); i++){
		p.drawPolygon(*parray[i]);
	}
}

void FreehandRegion::fetchScreen(QPixmap const &scr)
{
	// fetch Map Pixmap
	QImage img(scr.convertToImage());

	// avoid RED color
	int x,y;
	for(y=0; y<img.height(); y++){
		unsigned int *line = (unsigned int *)img.scanLine(y);
		for(x=0; x<img.width(); x++){
			if(((*line)&0xf8fcf8) == 0xf80000){
				*line = ((*line)&0xff000000) | 0xf00000;
			}
			line++;
		}
	}

	// convert to Pixmap
	saved_screen.convertFromImage(img);

	// draw help text
	QPainter p(&saved_screen);
	QColor gray(220,110,110);

	p.setPen(QPen(gray));
	p.setBrush(QBrush(gray));
	p.drawRect(0, 0, 59*8+4, 20);

	p.setPen(QPen(Qt::black));
	QFont font = QFont::defaultFont();
	font.setPixelSize(16);
	p.setFont(font);
	p.drawText(2, 16, QString::fromUtf8("フリーハンド範囲指定モード (D-モード終了, BS-Undo, OK-決定)"));
	p.end();

	// copy to canvas
	canvas.resize(scr.width(), scr.height());
	bitBlt(&canvas, QPoint(0,0), &saved_screen, saved_screen.rect());
}

void FreehandRegion::freeAll()
{
	saved_screen.resize(0,0);
	canvas.resize(0,0);
	points.resize(0);
	clearParray();
}

void FreehandRegion::hideWidget()
{
	parentWidget()->sendRequest("hide", "slider");
	parentWidget()->sendRequest("hide", "scale");
	parentWidget()->sendRequest("hide", "menu");
	parentWidget()->sendRequest("hide", "marker");
	parentWidget()->sendRequest("hide", "centermarker");
	parentWidget()->sendRequest("hide", "gpsview");
	parentWidget()->sendRequest("hide", "poiview");

	pMainWindow->disableMenu();
}

void FreehandRegion::showWidget()
{
	parentWidget()->sendRequest("show", "slider");
	parentWidget()->sendRequest("show", "scale");
	parentWidget()->sendRequest("show", "menu");
	parentWidget()->sendRequest("show", "marker");
	parentWidget()->sendRequest("show", "centermarker");
	parentWidget()->sendRequest("show", "gpsview");
	parentWidget()->sendRequest("show", "poiview");

	pMainWindow->enableMenu();
}

void FreehandRegion::mode_on()
{
	mode = FETCH;
	hideWidget();
	repaint();
}

void FreehandRegion::mode_off()
{
	freeAll();
	mode = NONE;
	showWidget();
	repaint();
}

int FreehandRegion::save_region()
{
	FreehandDialog dialog(parentWidget(), "hoge", TRUE);
	QString str;

	// for reduse memory comsumption
	saved_screen.resize(0, 0);

	dialog.setNowZoom(center.getZoom());
	int ret = dialog.exec();

	if(ret == QDialog::Rejected)
		return false;

	int min, max;
	min = dialog.getMinZoom();
	max = dialog.getMaxZoom();

	RegionMap *map;
	map = create_normalized_mask();

	// for reduse memory comsumption
	canvas.resize(0, 0);

	for(int i=max; i >= min; i--){
		QString data;
		data.sprintf("# zoom=%d\n", i);
		data += create_command(map, i, dialog.getFormat());
		data += "\n";
		str += data;
	}

	QClipboard *cb = QApplication::clipboard();
	cb->setText(str);

	((MainWidget *)parentWidget())->dispInfo(QString::fromUtf8("クリップボードに保存しました"));

	delete map;
	return true;
}

QString FreehandRegion::create_command(RegionMap *org_map,
									   int zoom, int format)
{
	QString str = "", line;
	QString lat, lon;
	RegionMap map(org_map->zoomTo(zoom));
	QArray<QRect> rarray;
	unsigned int i;

	rarray = map.getRects();
	for(i=0;i<rarray.size();i++){
		QRect &rect = rarray[i];

		lat = QString::number(rect.left());
		if(rect.width()>1)
			lat += ":" + QString::number(rect.right());

		lon = QString::number(rect.top());
		if(rect.height()>1)
			lon += ":" + QString::number(rect.bottom());

		if(format == FreehandDialog::ShellScript){
			line = "down_map\t%1\t%2\t%3\n";
		}else{
			line = "down_map(\"%1\",\t\"%2\",\t\"%3\");\n";
		}

		line = line.arg(zoom).arg(lat).arg(lon);
		str += line;
	}

	return str;
}

RegionMap *FreehandRegion::create_normalized_mask()
{
	int top, left;
	int offset_x, offset_y;
	int width, height;
	int zoom;

	zoom = center.getZoom()-8;

	/* get normalized topleft position */
	top = center.getY() - canvas.height()/2;
	left = center.getX() - canvas.width()/2;
	offset_y = top % IMAGE_SIZE;
	offset_x = left % IMAGE_SIZE;
	top -= offset_y;
	left -= offset_x;

	/* calculate normalized size */
	width = (offset_x + canvas.width() + IMAGE_SIZE-1)/IMAGE_SIZE;
	height = (offset_y + canvas.height() + IMAGE_SIZE-1)/IMAGE_SIZE;
	width *= IMAGE_SIZE;
	height *= IMAGE_SIZE;

	RegionMap *map = new RegionMap(left, top, width, height, zoom);
	QImage img(canvas.convertToImage());
	int x,y;

 	for(y = 0; y < img.height(); y++){
		unsigned int *line = (unsigned int *)img.scanLine(y);

		for(x = 0; x < img.width(); x++){
			if(((*line++)&0xf8fcf8) == 0xf80000){
				map->set(x + offset_x, y + offset_y);
			}
		}
	}

	return map;
}
