/*
  marker.h
  Copyright (C) 2008-2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <qlabel.h>
#include <qevent.h>
#include <qpen.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qfont.h>
#include <qregexp.h>
#include "mainwidget.h"
#include "marker.h"

#include <math.h>

Marker::Marker(LayeredBase *parent, const char *name) :
	LayeredWidget(parent, name)
{
	QString path;
	int i;
	conf = (QGConf *)lookupObject(this, "QGConf", "config");

	base.load(MARKER_FILE);

	for(i=0;i<9;i++){
		if(i==4) continue;

		path.sprintf(APP_PATH"/arrow%d.png", i);
		arrow[i].load(path);
	}
	for(i=0;i<12;i++){
		path.sprintf(APP_PATH"/arrow%02d.png", i);
		gps_arrow[i].load(path);
	}

	marker_on = conf->readNumEntry("MarkerOn", false);
	marker_bit = conf->readNumEntry("MarkerStat", 0);
	marker_timer_id = -1;
	marker_mode = conf->readNumEntry("MarkerMode", MAP);

	/* connect signals */
	connect(parent, SIGNAL(positionChanged(Pos &)),
			this, SLOT(setPosition(Pos &)));
	connect(parent, SIGNAL(configChanged(const QString &, const QString &)),
			this, SLOT(setConfig(const QString &, const QString &)));
}

Marker::~Marker()
{
}

bool Marker::paintEvent(QPaintEvent *e UNUSED, QPixmap &scr)
{
	int i;
	if(!marker_on)
		return false;

	for(i=0;i<10;i++){
		if(IS_MARKER_BIT(i)){
			drawMarker(&scr, saved_pos[i], i);
		}
	}
	return true;
}

void Marker::drawMarker(QPixmap *scr, Pos pos, int index)
{
	int sx,sy;
	int max_x = 256<<(19-center.getZoom());

	pos.zoomTo(center.getZoom());
	pos.normalize();

	sx = pos.getX()-center.getX();
	sy = pos.getY()-center.getY();

	if(sx < -max_x/2)
		sx+=max_x;
	else if(sx > max_x/2)
		sx-=max_x;

	sx += scr->width()/2;
	sy += scr->height()/2;

	if(isClip(scr, sx, sy)){
		if(marker_mode == MAP){
			drawClipMarker(scr, sx, sy, index);
		}else{
			drawClipGPSMarker(scr, sx, sy, index);
		}
		return;
	}

	drawBase(scr, sx, sy-8-5, index);
	QPainter p(scr);
	p.drawPixmap( sx-3, sy-7, arrow[7]);
}

void Marker::drawBase(QPixmap *scr, int cx, int cy, int index)
{
	QPainter p(scr);
#ifdef QWS
	QFont font("helvetica", 12, QFont::Bold);
#else
	QFont font("helvetica", 9, QFont::Bold);
#endif
	QString str;
	str = index;

	p.setFont(font);

	p.drawPixmap( cx-7, cy-8, base );

	str.sprintf("%d", index);
	p.drawText(cx-2, cy+4, str);
}

QPixmap &Marker::getArrow(int deg)
{
	deg/=10;
	deg=(deg+36)%12;
	return gps_arrow[deg];
}

bool Marker::isClip(QPixmap *scr, int sx, int sy)
{
	if(sy < M_TOP || sy > scr->height()-M_BOTTOM ||
	   sx < M_LEFT || sx > scr->width()-M_RIGHT)
		return true;

	return false;
}

void Marker::drawClipMarker(QPixmap *scr, int sx, int sy, int index)
{
	QPainter p(scr);
	int mx,my;
	int ax,ay;
	int top, bottom, left, right;
	int arrow;

	int offsety[] = {
		-M_ARROW_LEN45, -M_ARROW_LEN, -M_ARROW_LEN45,
		0, 0, 0,
		M_ARROW_LEN45, M_ARROW_LEN, M_ARROW_LEN45,
	};
	int offsetx[] = {
		-M_ARROW_LEN45, 0, M_ARROW_LEN45,
		-M_ARROW_LEN, 0, M_ARROW_LEN,
		-M_ARROW_LEN45, 0, M_ARROW_LEN45,
	};

	mx = my = ax = ay = 0;

	top = M_TOP + M_ARROW_LEN;
	bottom = scr->height() - M_BOTTOM - M_ARROW_LEN;
	left = M_LEFT + M_ARROW_LEN;
	right = scr->width() - M_RIGHT - M_ARROW_LEN;

	mx = ax = sx; my = ay = sy;
	arrow = 4;

	if(sy < top){
		my = top;
		arrow = 1;
	}else if(sy > bottom){
		my = bottom;
		arrow = 7;
	}

	if(sx < left){
		mx = left;
		arrow--;
	}else if(sx > right){
		mx = right;
		arrow++;
	}

	ay = my + offsety[arrow];
	ax = mx + offsetx[arrow];

	p.drawPixmap(ax-3, ay-3, this->arrow[arrow]);
	p.end();

	drawBase(scr, mx, my, index);
	
}

void Marker::drawClipGPSMarker(QPixmap *scr, int sx, int sy, int index)
{
	QPainter p(scr);
	int dx,dy;
	int mx,my;
	int ax,ay;
	int top, bottom, left, right;

	dx = sx-scr->width()/2;
	dy = sy-scr->height()/2;

	mx = my = ax = ay = 0;

	top = M_TOP - scr->height()/2;
	bottom = scr->height()/2 - M_BOTTOM;
	left = M_LEFT - scr->width()/2;
	right = scr->width()/2 - M_RIGHT;

	if(dx >= 0 && dy >= 0){
		getMarkerpos(dx, dy, right, bottom,
					 &mx, &my, &ax, &ay);
	}else if(dx >= 0 && dy < 0){
		getMarkerpos(dx, -dy, right, -top,
					 &mx, &my, &ax, &ay);
		my = -my;
		ay = -ay;
	}else if(dx < 0 && dy >= 0){
		getMarkerpos(-dx, dy, -left, bottom,
					 &mx, &my, &ax, &ay);
		mx = -mx;
		ax = -ax;
	}else if(dx < 0 && dy < 0){
		getMarkerpos(-dx, -dy, -left, -top,
					 &mx, &my, &ax, &ay);
		my = -my;
		mx = -mx;
		ay = -ay;
		ax = -ax;
	}

	int deg;
	deg = (int)(atan2(dx, dy)/3.1415926535*180.0);

	mx += scr->width()/2;
	my += scr->height()/2;
	ax += scr->width()/2;
	ay += scr->height()/2;

	QPen pen(Qt::black, 2);
	p.setPen(pen);
	p.drawLine(mx, my, ax, ay);

	p.drawPixmap(ax-3, ay-3, getArrow(deg));
	p.end();

	drawBase(scr, mx, my, index);
	
}

void Marker::getMarkerpos(int dx, int dy, int max_x, int max_y,
						  int *mx, int *my, int *ax, int *ay)
{
	int x, y;

	if(dx == 0){
		*ax = *mx = 0;
		*ay = max_y;
		*my = *ay - M_ARROW_LEN;
		return;
	}
	if(dy == 0){
		*ay = *my = 0;
		*ax = max_x;
		*mx = *ax - M_ARROW_LEN;
	}

	y = dy*max_x/dx;
	if(y < max_y){
		*ay = y;
		*ax = max_x;
		*mx = max_x-M_ARROW_LEN;
		*my = dy*(*mx)/dx;
		return;
	}

	x = dx*max_y/dy;
	*ax = x;
	*ay = max_y;
	*my = max_y-M_ARROW_LEN;
	*mx = dx*(*my)/dy;
}

void Marker::toggleMarker(void)
{
	marker_on = !marker_on;
	if(marker_on && !marker_bit){
		marker_bit=(1<<10)-1;
//		MARKER_TOGGLE(1);
	}

	conf->writeEntry("MarkerOn", marker_on);
	conf->writeEntry("MarkerStat", marker_bit);

	repaint();
	dispMarkerInfo();
}

void Marker::toggleMarker(int num)
{
	marker_on = true;

	MARKER_TOGGLE(num);
	if(!marker_bit)
		marker_on = false;

	conf->writeEntry("MarkerOn", marker_on);
	conf->writeEntry("MarkerStat", marker_bit);

	repaint();
	dispMarkerInfo();
}

bool Marker::keyPressEvent ( QKeyEvent * e )
{
	switch(e->key()){
	case Qt::Key_M:
		if(e->state() & ShiftButton){
			marker_mode = (marker_mode == GPS)?MAP:GPS;
			conf->writeEntry("MarkerMode", marker_mode);
			repaint();
			return true;
		}
		toggleMarker();
		return true;

	case Qt::Key_Q:
		if(marker_timer_id > 0){
			killTimer(marker_timer_id);
			marker_timer_id=-1;
		}
		((MainWidget *)parentWidget())->hideInfo();
		return true;
	case Qt::Key_0:
	case Qt::Key_1:
	case Qt::Key_2:
	case Qt::Key_3:
	case Qt::Key_4:
	case Qt::Key_5:
	case Qt::Key_6:
	case Qt::Key_7:
	case Qt::Key_8:
	case Qt::Key_9:
		if(marker_timer_id < 0)
			break;

		toggleMarker(e->key() - Qt::Key_0);
		return true;
	}
	return false;
}

void Marker::timerEvent( QTimerEvent *e )
{
	int id;

	id = e->timerId();
	if(id == marker_timer_id){
		killTimer(id);
		marker_timer_id=-1;
	}
}

void Marker::showEvent()
{
	Pos tmp;
	for(int i=0;i<=9;i++){
		QString str;
		str.sprintf("Pos%d", i);
		str = conf->readEntry(str, tmp.serialize());
		saved_pos[i].deserialize(str);
	}

	setGeometry(parentWidget()->rect());
}

void Marker::dispMarkerInfo()
{
	QString str;
	int i;

	if(marker_timer_id > 0){
		killTimer(marker_timer_id);
		marker_timer_id=-1;
	}

	if(marker_on){
		str = "MarkerON:";
		for(i=0;i<10;i++){
			str.sprintf("%s<font color=\"%s\">%d</font>",
						str.ascii(), IS_MARKER_BIT(i)?"#000":"#999", i);
		}
	}else{
		str = "MarkerOFF";
	}

	((MainWidget *)parentWidget())->dispInfo(str);
	marker_timer_id = startTimer(3000);
}

void Marker::setPosition(Pos &center)
{
	this->center = center;
}

void Marker::setConfig(const QString &key, const QString &value)
{
	int i;

	i = key.find(QRegExp("^Pos[0-9]$"));
	if( i == 0 ){
		i = key[3] - '0';
		saved_pos[i].deserialize(value);
		repaint();
	}
}

void Marker::request(QString const &req, QString const &arg)
{
	int ev = arg.toInt();

	if (req != "menu")
		return;

	if (ev == EV_TOGGLE_MARKER)
		toggleMarker();
	if (ev >= EV_TOGGLE_MARKER_0 && ev <= EV_TOGGLE_MARKER_9)
		toggleMarker(ev - EV_TOGGLE_MARKER_0);
}
