/*
  poi_add_dialog.cpp
  Copyright (C) 2008-2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include "qgmap.h"
#include "poi_add_dialog.h"
#include <qstring.h>
#include <qlabel.h>
#include <qvbox.h>
#include <qhbox.h>
#include <qpushbutton.h>
#include <qsizepolicy.h>
#include <qlayout.h>
#include <qlineedit.h>
#include <qpixmap.h>
#include <qmessagebox.h>
#include <qfileinfo.h>
#include <qcombobox.h>
#include <qspinbox.h>
#include <qlistbox.h>

#include "pos.h"
#include "filedialog.h"
#include "poi_db.h"
#include "poi_icon_dialog.h"
#include "poi_icon_select.h"

#define DIALOG_WIDTH  400
#ifdef QWS
#define DIALOG_HEIGHT (FRAME_HEIGHT*2+4)
#else
#define DIALOG_HEIGHT (FRAME_HEIGHT*2+40)
#endif

#define FRAME_HEIGHT 58

PoiAddDialog::PoiAddDialog(enum DialogMode mode, QWidget* parent,
						   const char* name, bool modal, WFlags fl)
	: QDialog( parent, name, modal, fl )
{
	this->mode = mode;
	db = new PoiDb(DB_POI_PATH);
	icon_id = -1;

	resize( QSize(DIALOG_WIDTH, 1).expandedTo(minimumSizeHint()) );

	if (mode == Add)
		setCaption(QString::fromUtf8("POI追加"));
	else
		setCaption(QString::fromUtf8("POI変更"));

//	setFont(QFont("lcfont", 16));

	QVBoxLayout *vbox = new QVBoxLayout(this, 0, 3);

	if (mode == Add) {
		QHBox *file_hbox = new QHBox(this);
		new QLabel(QString::fromUtf8("POI: "), file_hbox);
		file_edit = new QLineEdit(file_hbox);
		QPushButton *file_button = new QPushButton(QString::fromUtf8("参照"), file_hbox);
		vbox->addWidget(file_hbox);

		connect(file_button, SIGNAL(clicked()),
				this, SLOT(press_file_browse()));
	}

	QHBox *icon_hbox = new QHBox(this);
	new QLabel(QString::fromUtf8("Zoom:"), icon_hbox);
	zoom_spin = new QSpinBox(0, 19, 1, icon_hbox);
	zoom_spin->setValue(7);

	new QLabel(QString::fromUtf8("  Icon:"), icon_hbox);
	icon_button = new QPushButton(icon_hbox);
	vbox->addWidget(icon_hbox);

	QLabel *dummy = new QLabel(icon_hbox);
	dummy->setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding));

	QHBox *name_hbox = new QHBox(this);
	new QLabel(QString::fromUtf8("名前:"), name_hbox);
	name_edit = new QLineEdit(name_hbox);
	vbox->addWidget(name_hbox);

	connect(icon_button, SIGNAL(clicked()), this, SLOT(press_icon_button()));

	updateIcon();

#ifndef QWS
	QHBox *op_hbox = new QHBox(this);
	QPushButton *cancel = new QPushButton("Cancel", op_hbox);
	QPushButton *ok = new QPushButton("OK", op_hbox);
	ok->setDefault(true);
	vbox->addWidget(op_hbox);

	connect(cancel, SIGNAL(clicked()),
			this, SLOT(reject()));
	connect(ok, SIGNAL(clicked()),
			this, SLOT(accept()));
#endif
}

PoiAddDialog::~PoiAddDialog()
{
	delete db;
}

void PoiAddDialog::setId(int id)
{
	PoiDbFile *file = db->queryPoiFile(-1, id);

	if (!file)
		return;

	if (file->next()) {
		name_edit->setText(file->getName());
		icon_id = file->getIconId();
		zoom_spin->setValue(file->getZoom());
		updateIcon();
	}

	delete file;
}

QString PoiAddDialog::getPoiFile()
{
	return file_edit->text();
}

int PoiAddDialog::getIconId()
{
	return icon_id;
}

int PoiAddDialog::getZoom()
{
	return zoom_spin->text().toInt();
}

QString PoiAddDialog::getName()
{
	return name_edit->text();
}

void PoiAddDialog::accept()
{
	if (mode == Add) {
		QFileInfo fi(file_edit->text());

		if (!fi.isFile() || !fi.isReadable()) {
			QMessageBox::critical(this,
								  "Error!!",
								  QString::fromUtf8("POIファイルが開けません"));
			return;
		}
	}

	QDialog::accept();
}

void PoiAddDialog::press_file_browse()
{
	FileDialog dlg(this, QString::fromUtf8("Poiファイル"));
	dlg.setFileName(file_edit->text());
	if (dlg.exec() == QDialog::Accepted) {
		file_edit->setText(dlg.getFileName());
	}
}

void PoiAddDialog::press_icon_button()
{
	PoiIconSelect dlg(this);

	if (dlg.exec() == QDialog::Accepted) {
		icon_id = dlg.getIconId();
		updateIcon();
	}
}

void PoiAddDialog::updateIcon()
{
	if (icon_id >= 0) {
		icon_button->setPixmap(QPixmap(db->getPoiIcon(icon_id)));
		return;
	}

	PoiDbIcon *icon = db->queryPoiIcon();
	if (!icon)
		return;

	if (icon->next()) {
		icon_button->setPixmap(QPixmap(icon->getIconData()));
		icon_id = icon->getId();
	}

	delete icon;
}
